/*****************************************************************************
 * Project: BaBar detector at the SLAC PEP-II B-factory
 * Package: AFit
 * Author: A Bevan (a.j.bevan@qmul.ac.uk)
 *
 * Copyright (C) 2007 QMUL
 *****************************************************************************/

// -- CLASS DESCRIPTION [PDF] --
//
// Implementation of a Factory class to access multi-dimensional component 
// PDF models wrapped by an AFitMaster.  The following types are 
// available:
//
//  default
//  vvpolarisation
//
#include <iostream>
#include <fstream>

#include "TObjString.h"

#ifdef __BABARROOFIT__
#include "RooFitCore/RooAddPdf.hh"
#include "RooFitCore/RooArgSet.hh"
#include "RooFitCore/RooStringVar.hh"
#include "RooFitCore/RooDataSet.hh"
#include "RooFitCore/RooFitResult.hh"
#include "RooFitCore/RooSimultaneous.hh"
#include "RooFitModels/RooUnblindOffset.hh"
#include "RooFitCore/RooFormulaVar.hh"
#else
#include "RooAddPdf.h"
#include "RooStringVar.h"
#include "RooArgSet.h"
#include "RooDataSet.h"
#include "RooFitResult.h"
#include "RooSimultaneous.h"
#include "RooUnblindOffset.h"
#include "RooFormulaVar.h"
#endif

/*
 * The following includes are AFit PDFs 
 */
#include "AFit/AFitMaster.hh"
#include "AFit/AFitComponentFactory.hh"

using namespace std;

AFitMaster::AFitMaster(TString configfile) :
  _configFile(configfile)
  , discVars("variables", "", "null")
  , condVars("conditional_variables", "", "null")
  , components("components", "", "null")
  , fitOptions("fitOptions", "", "ertm")
  , catVars("catVars", "", "null")
  , splitRule("splitRule", "", "null")
  , discVarSet("Discriminating_Variable_Set")
  , condVarSet("Conditional_Variable_Set")
  , compVarSet("Component_Variable_Set")
  , coefSet("Fit_Coefficient_Set")
  , configSet("Configuration_Set")
  , compList("Fit_Component_List")
  , fin(0)
{
  setVerbose();
  if(_debug) cout << ClassName() << _ctorstr << endl;
  thePdf = 0;
  pdfBld = 0;
  bldConfig = 0;
  theSimPdf = 0;

  configSet.add(discVars);
  configSet.add(condVars);
  configSet.add(components);
  configSet.add(fitOptions);
  configSet.add(catVars);
  configSet.add(splitRule);
  readConfigFile(_configFile);
}

AFitMaster::~AFitMaster() {
}

void AFitMaster::readConfigFile(TString cfgFile) {
  /*
   * read in the fit configuration, and construct the list of 
   * discriminating variables
   */
  configSet.readFromFile(cfgFile, 0, "FitConfiguration");

  // create and read in discriminating variables
  TString tmpStr = discVars;
  TObjArray * discVarArr = tmpStr.Tokenize(',');
  for(Int_t i=0; i < discVarArr->GetEntries(); i++) {
    TString thisVar = ((TObjString*)discVarArr->At(i))->GetString();
    Double_t min(0.0);
    Double_t max(1.0);
    RooRealVar * newVar = new RooRealVar( thisVar , "Discriminating Variable", min, max);

    discVarSet.add(*newVar);
  }
  discVarSet.readFromFile(_configFile, 0, "FitConfiguration");

  // create and read in conditional variables 
  TString tmpStr2 = condVars;
  if(tmpStr2 != "null") {
    TObjArray * condVarArr = tmpStr2.Tokenize(',');
    for(Int_t i=0; i < condVarArr->GetEntries(); i++) {
      TString thisVar = ((TObjString*)condVarArr->At(i))->GetString();
      Double_t min(0.0);
      Double_t max(1.0);
      RooRealVar * newVar = new RooRealVar( thisVar , "Conditional Variable", min, max);
      
      condVarSet.add(*newVar);
    }
    condVarSet.readFromFile(_configFile, 0, "FitConfiguration");
  }

  // create and read in categories 
  TString tmpStr3 = catVars;
  if(tmpStr3 != "null") {
    TObjArray * catArr = tmpStr3.Tokenize(',');
    for(Int_t i=0; i < catArr->GetEntries(); i++) {
      TString thisCat = ((TObjString*)catArr->At(i))->GetString();
      RooCategory * newCat = new RooCategory( thisCat , "Category Variable");
      
      // now read the category information from the file
      // and define all category types for this variable 
      RooStringVar thisTmpVar(thisCat + "_categories", "", "null");
      RooArgSet thisTmpSet(thisTmpVar);
      thisTmpSet.readFromFile(_configFile, 0, "FitConfiguration");
      TString catLabels = thisTmpVar;
      TObjArray * catLabelArr = catLabels.Tokenize(',');
      for(Int_t j=0; j < catLabelArr->GetEntries(); j++) {
	TString type = ((TObjString*)catLabelArr->At(j))->GetString();
	newCat->defineType( type );
      }
      catSet.add(*newCat);

      TString thesplitRule = splitRule;
      thesplitRule.ReplaceAll("null", "");
      if( !thesplitRule.Contains( newCat->GetName() ) ) {
	// the category split rule has not been specified, so check the data card
	// for the rule under '<category name>_splitrule'
	RooStringVar thisSplitRule(TString(newCat->GetName())+"_splitrule", "Split Rule for this category", "null");
	RooArgSet thisTmpSet(thisSplitRule);
	thisTmpSet.readFromFile(_configFile, 0, "FitConfiguration");
	thesplitRule += " ";
	thesplitRule += thisSplitRule;
	splitRule = thesplitRule;
      }
    }
  }

  if(_debug) {
    cout << "____________________________________________________" << endl;
    cout << ClassName() << " have created the following discriminating variables" << endl;
    discVarSet.Print("v");
    if(condVarSet.getSize()) {
      cout << "____________________________________________________" << endl;
      cout << ClassName() << " have created the following conditional variables" << endl;
      condVarSet.Print("v");
    }

    if(catSet.getSize()) {
      cout << "____________________________________________________" << endl;
      cout << ClassName() << " have created the following category variables" << endl;
      catSet.Print("v");
      
      TString mySR = splitRule;
      cout << "____________________________________________________" << endl;
      cout << "the split rule to use is " << mySR << endl;
      cout << "____________________________________________________" << endl;
    }
  }
}

RooAbsPdf * AFitMaster::getPdf() {
  if(!thePdf) thePdf = makePdf();
  return thePdf;
}

RooAbsPdf * AFitMaster::getSimPdf() {
  // first of all make the nominal PDF, prior to applying the rules for 
  // splitting the PDF according to categories specified in the datacard.
  if(!thePdf) thePdf = makePdf();

  /*
  // now apply the splitting rules and make the RooSimPdfBuilder
  // are there any specializations to make?
  RooArgSet Specializations("Specializations");
  */

  // the pdf builder
  pdfBld = new RooSimPdfBuilder(*thePdf);
  bldConfig = ((RooSimPdfBuilder*)pdfBld)->createProtoBuildConfig();

  // set the physics model name to that of the original pdf
  RooStringVar * a = (RooStringVar *)bldConfig->find("physModels");
  a->setVal(thePdf->GetName());

  // specify the category to split by
  RooStringVar * splitCatsRSV = (RooStringVar *)bldConfig->find("splitCats");
  TString splitCats = catVars;
  splitCats.ReplaceAll(",", " ");
  splitCatsRSV->setVal(splitCats);

  // set the list of parameters to split by the category
  RooStringVar * splitRuleRSV = (RooStringVar *)bldConfig->find(thePdf->GetName());
  splitRuleRSV->setVal( splitRule.getVal() );

  //  if(Specializations.getSize()) pdfBld->addSpecializations(Specializations);

  RooArgSet varSet;
  varSet.add(discVarSet);
  varSet.add(condVarSet);
  varSet.add(catSet);
  RooDataSet * dummySet = new RooDataSet("dummySet", "dummySet", varSet);

  cout << "____________________________________________________________" << endl;
  const RooSimultaneous* tmppdf = dynamic_cast<const RooSimultaneous*>
    (((RooSimPdfBuilder*)pdfBld)->buildPdf(*bldConfig, dummySet));

  cout << "____________________________________________________________" << endl;
  assert(tmppdf);
  theSimPdf = new RooSimultaneous(*tmppdf);
  theSimPdf->SetName("TotalPdf_Simultaneous");
  theSimPdf->SetTitle("Total categorized Pdf");

  // now read the configuration file to pick up values for each 
  // of the split paramters
  RooArgSet * simCompSet = theSimPdf->getComponents();
  RooArgSet * simParSet  = theSimPdf->getParameters(simCompSet);
  simParSet->readFromFile(_configFile);

  return theSimPdf;
}


RooAbsPdf * AFitMaster::makePdf() {
  cout << "----------------------------------------------------------------------------" << endl;
  cout << ClassName() << "::makePdf(): Will construct the fit model" << endl;
  cout << "----------------------------------------------------------------------------" << endl;
  RooAbsPdf * thePdf(0);

  AFitAbsPdfBuilder::setDataCard(_configFile);
  cout << ClassName() << "RooAbsPdf * AFitMaster::makePdf(TString configfile) to implement" << endl;
  /*
   * Tokenize the list of components, and then make them one by one 
   */
  TString tmpStr = components;
  TObjArray * compArr = tmpStr.Tokenize(',');
  for(Int_t i=0; i < compArr->GetEntries(); i++) {
    TString thisComp = ((TObjString*)compArr->At(i))->GetString();
    RooStringVar * thisVar = new RooStringVar(thisComp, "", "null");
    compVarSet.add(*thisVar);
  }
  compVarSet.readFromFile(_configFile, 0, "FitConfiguration");

  if(_debug) {
    cout << ClassName() << " : Will construct a fit model with the following components" << endl;
    compVarSet.Print("v");
  }

  // ensure that factories are informed about conditional variables
  if(condVarSet.getSize()) compFactory.setConditionalVariables(condVarSet);
  compFactory.setConfigFile(_configFile);

  /*
   * Now loop over the components and build the PDF model
   */
  TIterator * iter = compVarSet.createIterator();
  TObject * obj;
  while(obj = iter->Next()) {
    RooStringVar &thisCompVar = *(RooStringVar*)obj;
    RooArgList tmpList(discVarSet);

    RooAbsPdf * thisComponentPdf = compFactory.makeComp(thisCompVar.GetName(), thisCompVar.getVal(), tmpList );
    compList.addOwned(*thisComponentPdf);

    // make the coefficient to go with this component.  The coefficient
    // is a signal yield, multiplied by a tagging efficiency, and 
    // has the option of being blind or unblind.  The tagging efficiency
    // is 1 by default, and can be specified as a split cat in the case
    // that one is performing a simultaneous fit without needing to
    // fit for individual yields.

    // 1. The yield
    TString coeffName = thisCompVar.GetName();    coeffName += "Yield";
    RooRealVar * thisCoefficient = new RooRealVar (coeffName, "coefficient", 0.0);
    coefSet.addOwned(*thisCoefficient);

    // 2. The tagging efficiency
    TString coeffEffName = thisCompVar.GetName();    coeffEffName += "_tageff";
    RooRealVar * thisEffCoefficient = new RooRealVar (coeffEffName, "tagging efficiency coefficient", 1.0);
    coefSet.addOwned(*thisEffCoefficient);

    // 3. The tagging efficiency * yield
    TString coeffYName = thisCompVar.GetName();    coeffYName += "_yield_times_tageff";
    RooFormulaVar * thisCoefYield = new RooFormulaVar(coeffYName, "tagged yield", "@0*@1", RooArgSet(*thisEffCoefficient, *thisCoefficient));
    coefSet.addOwned(*thisCoefYield);

    // Blinding of the yield parameter
    RooCategory * BlindingType = new RooCategory(coeffName + "_BlindingType", "is the yield Blind or UnBlind");
    BlindingType->defineType("blind",1);
    BlindingType->defineType("unblind",0);
    BlindingType->setLabel("unblind");
    coefSet.addOwned(*BlindingType); 

    RooStringVar * _thisBlindString = new RooStringVar(coeffName + "_BlindingString", "yield blinding string", "sixNations");
    coefSet.addOwned(*_thisBlindString);

    RooRealVar * _thisBlindScale = new RooRealVar(coeffName + "_BlindingScale", "yield blinding scale", 200.0);
    coefSet.addOwned(*_thisBlindScale);

    // 4. The unblind/blind [tagging efficiency * yield]
    TString ubCoeffName = thisCompVar.GetName();  ubCoeffName += "UnBlindYield";
    RooUnblindOffset * UnBlindYield = new RooUnblindOffset(ubCoeffName, "unblind yield",
							   _thisBlindString->getVal(), _thisBlindScale->getVal(), *thisCoefYield, *BlindingType);

    ubCoefSet.addOwned(* UnBlindYield);
  }
  coefSet.readFromFile(_configFile, 0, "FitConfiguration");
  ubCoefSet.readFromFile(_configFile, 0, "FitConfiguration");
  cout << " -------> COEFSET READ FROM FILE " << endl;

  RooArgList coefList(ubCoefSet); // use the (ub)blind yields as coefficients in the extended fit
  if(_debug) {
    cout << endl << ClassName() << " : Making a PDF with the following components and coeffients:" << endl;
    cout << "  NAME " << setw(30) << " YIELD " << setw(38) << " COMPONENT TYPE" << endl;
    RooArgList tmpList(compVarSet);
    for(Int_t i=0; i < compList.getSize(); i++) {
      RooAbsReal   & thisCoefVar = (RooAbsReal&)coefList[i];
      RooStringVar & thisCompVar = (RooStringVar&)tmpList[i];
      cout << setw(30) << setiosflags(ios::left) << thisCompVar.GetName() << " " << setw(30) << thisCoefVar.getVal() << " " << setw(30) << thisCompVar.getVal()<< resetiosflags(ios::left) << endl;
    }
  }

  thePdf = new RooAddPdf("TotalModel", "Fit Master Model", compList, coefList);

  getDiscVarSet().Print("v");
  getCoefSet().Print("v");
  getUbCoefSet().Print("v");

  return thePdf;
}



void AFitMaster::fitParameters(ostream & out) {
  /*
   * Given a pdf, one by one, loop over the fit components
   * and fit the PDF to reference data.  If a component PDF
   * is uncorrelated, then the PDF is fit to data one 
   * discriminating variable at a time, otherwise, all
   * parameters are fitted simultaneoulsy. 
   *
   * The argument parsed to this function is the stream
   * that the configuration details of the fit model
   * should be written to once the fit parameters have
   * been determined from the reference data.
   *
   * The datacard block labeled with [PDF Param Files] is used
   * to read in the options for each component.  The information
   * contained in this block has the following format:
   *
   *  'componentname'_referencedata      = 'somevalue'
   *  'componentname'_referencedata_long = 'somevalue'
   *  'componentname'_referencedata_tran = 'somevalue'
   *
   * where the first example is for a default component type,
   * and the second and third exampls are for a vvpolarisation
   * component type, with longitudinal, and transverse polarsiations,
   * respectively.
   *
   * The value of the string assigned to these variables is
   * a comma separated list, with at least one entry.  The elements
   * of this list are:
   *
   *   datafile,datatype,dataname,correlated
   *
   * where
   *
   *   datafile            is the name of the data file to fit the 
   *                       PDF component to.
   *   datatype            is the type of data to read in. This is either
   *                         ascii
   *                         ttree
   *                         RooDataSet (rds for short)  
   *   correlated          a flag specifying if the component PDF
   *                       is made up of uncorrelated 1D PDFs.  This
   *                       can take the values
   *                         yes
   *                         no
   *
   */
  cout << "----------------------------------------------------------------------------" << endl;
  cout << ClassName() << "::fitParameters(ostream & out): Will fit PDF parameters for all fit components" << endl;
  cout << "----------------------------------------------------------------------------" << endl;
  TString fitParameterBlockName = "PDF Param Files";
  // ensure that a PDF exists
  if(!thePdf) makePdf();
  if(!thePdf) {
    cout << ClassName() << "::fitParameters ERROR  - unable to make the TotalPDF - quitting" << endl;
    exit(1);
  }
  // now check for availability of data samples

  compList.Print("v");

  RooArgList tmpList(compVarSet);
  for(Int_t i=0; i < compList.getSize(); i++) {
    RooStringVar & thisCompVar = (RooStringVar&)tmpList[i];
    cout << ClassName() << "::fitParameters - fitting PDF parameters for" << endl;
    cout << "  NAME = " << thisCompVar.GetName() << " COMPONENT TYPE = " << thisCompVar.getVal() << endl;
    
    TString thisCompVarVal = thisCompVar.getVal();
    if(thisCompVarVal.Contains( "default") ){
      TString datafilename = thisCompVar.GetName();
      datafilename += "_referencedata";
     
      /*
       * now have to retrieve data, and fit the pdf component to the data
       */
      Bool_t kCorrelatedPDFs=kTRUE;
      RooDataSet * componentData = getDataForFitting(datafilename, fitParameterBlockName, kCorrelatedPDFs);

      cout << ClassName() << ": The data set from " << datafilename << " contains " << componentData->numEntries() << " entries" << endl;
      componentData->Print("v");
      fitPDFToData(thisCompVar.GetName(), componentData, kCorrelatedPDFs);

    } else if(thisCompVarVal.Contains( "vvpolarisation") ){
      TString datafilename_long = thisCompVar.GetName();
      datafilename_long += "_referencedata_long";
      cout << datafilename_long << endl;
      TString datafilename_tran = datafilename_long;
      datafilename_tran.ReplaceAll("long", "tran");
      cout << datafilename_tran << endl;
      /*
       * now have to retrieve data, and fit the pdf component to the data
       */
      Bool_t kCorrelatedPDFsL=kTRUE;
      RooDataSet * componentDataLong = getDataForFitting(datafilename_long, fitParameterBlockName, kCorrelatedPDFsL);
      cout << ClassName() << ": The data set from " << datafilename_long << " contains " << componentDataLong->numEntries() << " entries" <<endl;
      componentDataLong->Print("v");

      TString lname = thisCompVar.GetName(); lname +="_long";
      fitPDFToData(lname, componentDataLong, kCorrelatedPDFsL);

      Bool_t kCorrelatedPDFsT=kTRUE;
      RooDataSet * componentDataTran = getDataForFitting(datafilename_tran, fitParameterBlockName, kCorrelatedPDFsT);
      cout << ClassName() << ": The data set from " << datafilename_tran << " contains " << componentDataTran->numEntries() << " entries" <<endl;
      componentDataTran->Print("v");

      TString tname = thisCompVar.GetName(); tname +="_tran";
      fitPDFToData(tname, componentDataTran, kCorrelatedPDFsT);
    } else {
      cout << ClassName() << "::fitParameters ERROR Component " << thisCompVar.GetName() << " has an unknown type =  " << thisCompVar.getVal() << endl;
    }
  }

  // fix PDF parameters for all PDFs before writing the configuration to out
  RooArgSet * compos = thePdf->getComponents();
  RooArgSet * pdfpar = thePdf->getParameters(compos);

  TIterator * iter = pdfpar->createIterator();
  TIter next(iter);
  TObject * obj(0);
  while (obj = (TObject *)next()) {
    RooRealVar * var = dynamic_cast<RooRealVar*>(obj);
    if(var) var->setConstant(kTRUE);
  }

  writeDataCard(out);
}


void AFitMaster::writeDataCard(ostream & out) {
  out << "// __________________________________________________________" << endl;
  out << "// " << endl;
  out << "// The following section lists the global fit configuration " << endl;
  out << "// " << endl;
  out << "// __________________________________________________________" << endl;
  out << "[FitConfiguration]" << endl;
  out << "// specify the variables and components to use in the fit" << endl;
  out << "variables = " << discVars.getVal() << endl;
  out << "components = " << components.getVal() << endl;
  out << "fitOptions = " << fitOptions.getVal() << endl;
  out << endl ;

  discVarSet.writeToStream(out, 0, "FitConfiguration");
  out << endl ;
  compVarSet.writeToStream(out, 0, "FitConfiguration");
  out << endl ;
  coefSet.writeToStream(out, 0, "FitConfiguration");
  out << endl << endl;

  // output a configuration file for the model
  compFactory.writeConfigFile(out);
}



RooDataSet * AFitMaster::getDataForFitting(TString filevarname, TString blockname, Bool_t &kCorrelated) {
  /*
   * get the data from the file specified by the filevarname, variable in the 
   * configuration card.  This variable is a comma separated list:
   *
   *  filename,dataType,dataName,correlatedPDFs
   *
   *  filename = physical filename
   *  dataType = ascii, tree or RooDataSet         (default is dataname)
   *  dataName = key name of data in a ROOT file   (default is null)
   *  correlatedPDFs = correlated or uncorrelated  (default is correlated)
   *
   */
  RooStringVar fName(filevarname, "fname", "");
  RooArgSet aset;
  aset.add(fName);
  aset.readFromFile(_configFile,0,"PDF Param Files") ;

  TString fname = fName.getVal();
  TString dataName = "dataname";
  TString dataType = "null";
  TString correlatedPDFs = "correlated";

  TObjArray * dataFileVarArr = fname.Tokenize(',');
  if(dataFileVarArr->At(0)) fname          = ((TObjString*)dataFileVarArr->At(0))->GetString();
  if(dataFileVarArr->At(1)) dataType       = ((TObjString*)dataFileVarArr->At(1))->GetString();
  if(dataFileVarArr->At(2)) dataName       = ((TObjString*)dataFileVarArr->At(2))->GetString();
  if(dataFileVarArr->At(3)) correlatedPDFs = ((TObjString*)dataFileVarArr->At(3))->GetString();
  dataType.ToLower();
  if(correlatedPDFs.Contains("uncorrelated")) kCorrelated=kFALSE;

  ifstream in(fname);
  if(!in) {
    cout << ClassName() <<"::getDataForFitting ERROR - unable to open " << fname <<endl;
    return 0;
  }

  if(fin) fin->Close();
  RooDataSet * data(0);
  if(dataType.Contains("ascii")) {
    data = RooDataSet::read(fname,discVarSet,"q","") ;
  } else if(dataType.Contains("roodataset")||dataType.Contains("rds")) {
    fin = new TFile(fname);
    data = (RooDataSet*)fin->Get(dataName); // assume that the data is a RooDataSet
  } else if(dataType.Contains("tree")){
    fin = new TFile(fname);
    TTree * tree = (TTree*)fin->Get(dataName);
    if(tree) {
      data = new RooDataSet(dataName, "data from "+fname, tree, discVarSet);
    } else {
      cout << ClassName() <<"::getDataForFitting ERROR - unable retrieve tree " << dataName << " from file " << fname <<endl;
      return 0;
    }
  }

  if(!data) {
    cout << ClassName() <<"::getDataForFitting ERROR - unable retrieve data " << dataName << " from file " << fname << "(type = " << dataType << ")" << endl;
    return 0;
  } return data;
}

void AFitMaster::fitPDFToData(TString pdfComponentName, RooDataSet * thisData, Bool_t &kCorrelated){
  if(!thisData) return;
  cout << ClassName() << "::fitPDFToData" << endl;
  cout << "\tFitting component " << pdfComponentName << " to data from " << thisData->GetName() << " (" << thisData->numEntries() << " data to fit)" << endl;
  if(kCorrelated) cout << "\tFit variables are assumed to be correlated" << endl;
  else cout << "\tFit variables are assumed to be uncorrelated" << endl;

  const RooArgSet * aset = thePdf->getComponents();
  RooAbsPdf *componentPdf = (RooAbsPdf*)aset->find(pdfComponentName);
  if(!componentPdf) { 
    cout << ClassName() << ":fitPDFToData CONFIGURATION ERROR - unable to locate " << pdfComponentName << endl;
    return;
  }

  Int_t Ndata = thisData->numEntries();
  if(Ndata==0) return;

  TString fitOpt = fitOptions.getVal();
  if(componentPdf->extendMode() == RooAbsPdf::CanNotBeExtended) fitOpt.ReplaceAll("e", "");
  cout << "still here" << endl;

  if(kCorrelated) {
    RooFitResult * r(0);
    r = componentPdf->fitTo(*thisData, fitOpt);
    if(r) {
      r->SetName(pdfComponentName+"_fitResult");
      resultList.Add(r);
    }
  } else {
    /*
     * fit one dimension at a time if the PDFs are uncorrelated.  This
     * is a lot faster than fitting all dimensions simultaneously.
     */
    RooArgList tmpList(discVarSet);
    for(Int_t i=0; i < discVarSet.getSize(); i++) {
      RooFitResult * r(0);
      cout << "i = " << i << endl;
      RooAbsReal   & thisDiscVar = (RooAbsReal&)tmpList[i];
      TString subComponentPDFName = componentPdf->GetName();
      subComponentPDFName += "_";
      subComponentPDFName += thisDiscVar.GetName();

      const RooArgSet * aset = thePdf->getComponents();
      RooAbsPdf *subComponentPdf = (RooAbsPdf*)aset->find(subComponentPDFName);
      cout << "going to fit again! "<< subComponentPDFName << " " << subComponentPdf << " data= " << thisData << " opt = " << fitOpt <<endl;
      if(!subComponentPdf) return;

      subComponentPdf->Print("v");
      r = subComponentPdf->fitTo(*thisData, fitOpt);
      if(r) {
	r->SetName(subComponentPDFName+"_fitResult");
	resultList.Add(r);
      }
    }
  } 

  return;
}

RooFitResult * AFitMaster::fitData(RooDataSet * thisData) {
  /*
   * Fit the total model to data
   */
  cout << "----------------------------------------------------------------------------" << endl;
  cout << ClassName() << "::fitData(RooDataSet * thisData): Will fit to the data" << endl;
  cout << "----------------------------------------------------------------------------" << endl;
  if(!thisData) {
    cout << ClassName() << "::fitData(RooDataSet * thisData) ERROR - specified data set is null" << endl;
    return 0;
  }
  if(!thePdf) thePdf = makePdf();
  RooFitResult * r = thePdf->fitTo(*thisData, fitOptions.getVal());
  return r;
}

void AFitMaster::persist(TString ofileName) {
  cout << ClassName() << ":Persisting the PDF model fit parameter information" << endl;
  cout << ClassName() << ":Output file is " << ofileName << endl;

  TFile f(ofileName, "RECREATE");
  f.cd();
  discVarSet.Write();
  compVarSet.Write();
  coefSet.Write();
  configSet.Write();
  resultList.SetName("PDF_Param_Fit_Result_List");
  resultList.Write();
  f.Write();
  f.Close();
}
