/*****************************************************************************
 * Project: BaBar detector at the SLAC PEP-II B-factory
 * Package: AFit
 *    File: $Id: AFitResolution.cc,v 1.7 2006/05/22 19:53:16 bevan Exp $
 * Authors:
 *   P-F Giraud, CEA/Saclay
 *   J Stark, IN2P3
 *   Ch Yeche, CEA/Saclay
 *
 * History:
 *   - adding MC truth model (Andreas, 11/22/2002)
 *   - migrated from RptBaseTools to AFit (Adrian, 05/11/2008)
 *
 * Copyright (C) 2002 CEA/Saclay, IN2P3
 *****************************************************************************/

// -- CLASS DESCRIPTION [AFit] --
// Builder for the resolution function

#include "AFit/AFitResolution.hh"
#ifdef __BABARROOFIT__
#include "RooFitModels/RooGaussModel.hh"
#include "RooFitCore/RooAddModel.hh"
#include "RooFitCore/RooRealConstant.hh"
#include "RooFitCore/RooTruthModel.hh"
#else
#include "RooGaussModel.h"
#include "RooAddModel.h"
#include "RooRealConstant.h"
#include "RooTruthModel.h"
#endif

ClassImp(AFitResolution)
  ;

AFitResolution::AFitResolution(RooAbsReal * var, RooAbsReal * varErr, TString name)
  : 
  sigMean1(name+TString("_sigMean1"), "Signal mean1 ",-0.128, -2.0, 2.0),
  sigMean2(name+TString("_sigMean2"), "Signal mean2 ",-0.403, -3.0, 3.0),
  sigMean3(name+TString("_sigMean3"), "Signal mean3 ",0),
  sigSig1(name+TString("_sigSig1"), "Signal sigma 1 ",0.666, 0.0, 2.0),
  sigSig2(name+TString("_sigSig2"), "Signal sigma 2 ",2.166, 0.0, 5.0),
  sigSig3(name+TString("_sigSig3"), "Signal sigma 3 ",8),
  coreFrac(name+TString("_coreFrac"), "Signal Core Fraction ",0.739, 0.0, 1.0),
  tailFrac(name+TString("_tailFrac"), "Signal Tail fraction ",0.01, 0.0, 1.0) ,
  scaleTailWidth(name+TString("_scaleTailWidth"), "Use per-event errors to scale tail widths ?"),
  scaleTailMean(name+TString("_scaleTailMean"), "Use per-event errors to scale tail offsets ?"),
  scaleCoreWidth(name+TString("_scaleCoreWidth"), "Use per-event errors to scale core widths ?"),
  scaleCoreMean(name+TString("_scaleCoreMean"), "Use per-event errors to scale core offsets ?"),
  useTruthModel(name+TString("_useTruthModel"), "Use Truth Resolution Model ?")
{
  _name = name;
  _var = var;
  _varErr = varErr;

  ownResolution = false;
  sigGauss1=0;
  sigGauss2=0;
  sigGauss3=0;

  scaleTailWidth.defineType("yes");
  scaleTailWidth.defineType("no");
  scaleTailWidth.setLabel("no");

  scaleTailMean.defineType("yes");
  scaleTailMean.defineType("no");
  scaleTailMean.setLabel("no");

  scaleCoreMean.defineType("yes");
  scaleCoreMean.defineType("no");
  scaleCoreMean.setLabel("no");

  scaleCoreWidth.defineType("yes");
  scaleCoreWidth.defineType("no");
  scaleCoreWidth.setLabel("no");

  useTruthModel.defineType("yes");
  useTruthModel.defineType("no");
  useTruthModel.setLabel("no");

  varSet.add(sigSig1);
  varSet.add(sigMean1);
  varSet.add(sigSig2);
  varSet.add(sigMean2);
  varSet.add(sigSig3);
  varSet.add(sigMean3);
  varSet.add(coreFrac);
  varSet.add(tailFrac);
  varSet.add(scaleCoreMean);
  varSet.add(scaleCoreWidth);
  varSet.add(scaleTailWidth);
  varSet.add(scaleTailMean);

  varSet.add(useTruthModel);

  readDataCard();
}

AFitResolution::~AFitResolution() {
  if (ownResolution) {
    globalConfig.remove(*_thePdf);
    if (!(useTruthModel == "yes")) {
      if(sigGauss1) delete sigGauss1;
      if(sigGauss2) delete sigGauss2;
      if(sigGauss3) delete sigGauss3;
    }
  }
}

void AFitResolution::buildPdf() {
  cout<< ClassName() <<  " Building a resolution function for " << _var->GetName() << " called "<< _name << endl;
  if (_varErr) _varErr->Print(); 
  else cout << "Conditional Varible " << _var->GetName() << " error is not being used" << endl; 
  sigSig1.Print();
  sigMean1.Print();
  sigSig2.Print();
  sigMean2.Print();
  sigSig3.Print();
  sigMean3.Print();
  coreFrac.Print();
  tailFrac.Print();
  cout<<"\tscaleCoreMean      = "<< scaleCoreMean.getLabel()<<endl;
  cout<<"\tscaleCoreWidth     = "<< scaleCoreWidth.getLabel()<<endl;
  cout<<"\tscaleTailMean      = "<< scaleTailMean.getLabel()<<endl;
  cout<<"\tscaleTailWidth     = "<< scaleTailWidth.getLabel()<<endl;
  cout<<"\tscaleTruthModel    = "<< useTruthModel.getLabel()<<endl;
 
  cout << "Checking the globalConfig for the resolution model\n";
  _thePdf = dynamic_cast<RooResolutionModel*>(globalConfig.find(_name));
  if(!_thePdf) cout << "  model does not exist - will construct" << endl;

  if (!_thePdf) {
    cout << "going to build resolution\n";
    ownResolution = true;
    if (useTruthModel == "yes") {
      cout << "Using truth model" << endl; 
      _thePdf = new RooTruthModel( _name + TString("truthModel"), "Truth Resolution Model", *(RooRealVar*)_var );
    }
    else {
      cout << "Using triple Gaussian model" << endl; 
      if(!_varErr) {
	if((scaleTailMean == "yes")  || (scaleCoreMean == "yes") || (scaleTailWidth == "yes") || (scaleCoreWidth == "yes") ) {
	  cout << "Not using the conditional variable in the resolution function " << _name << " - will not scale mean and width" << endl;
	}
	scaleCoreMean.setLabel("no");
	scaleTailMean.setLabel("no");
	scaleCoreWidth.setLabel("no");
	scaleTailWidth.setLabel("no");
      }

      sigGauss1 = new RooGaussModel(_name+TString("sigGaussCore_"), "Signal gauss1",
				    *(RooRealVar*)_var, sigMean1, sigSig1, 
				    (scaleCoreMean == "yes") ? *(RooRealVar*)_varErr : (RooRealVar&)RooRealConstant::value(1),
				    (scaleCoreWidth == "yes") ? *(RooRealVar*)_varErr : (RooRealVar&)RooRealConstant::value(1));
	
      sigGauss2 = new RooGaussModel(_name+TString("sigGaussTail_"), "Signal gauss2",
				    *(RooRealVar*)_var, sigMean2, sigSig2,
				    (scaleTailMean == "yes") ? *(RooRealVar*)_varErr : (RooRealVar&)RooRealConstant::value(1),
				    (scaleTailWidth == "yes") ? *(RooRealVar*)_varErr : (RooRealVar&)RooRealConstant::value(1));

      // _never_ use per-event errors for the third gaussian
      sigGauss3 = new RooGaussModel(_name+TString("sigGaussOutlier_"), "Signal gauss3",
                                    *(RooRealVar*)_var, sigMean3, sigSig3);

      Bool_t flag(kTRUE);
      ((RooGaussModel*)sigGauss1)->advertiseAymptoticIntegral(flag);
      ((RooGaussModel*)sigGauss1)->advertiseFlatScaleFactorIntegral(flag);
      ((RooGaussModel*)sigGauss2)->advertiseAymptoticIntegral(flag);
      ((RooGaussModel*)sigGauss2)->advertiseFlatScaleFactorIntegral(flag);
      ((RooGaussModel*)sigGauss3)->advertiseAymptoticIntegral(flag);
      ((RooGaussModel*)sigGauss3)->advertiseFlatScaleFactorIntegral(flag);

      _thePdf = new RooAddModel(_name, "resolution function",
                                RooArgList(*sigGauss1,*sigGauss2,*sigGauss3),
                                RooArgList(coreFrac,tailFrac));
      
      globalConfig.add(*_thePdf);
    }
  }
}

